#
#  dna : A Library of Procedures for Analysis of DNA Sequences
#  Copyright (C) 2000 J.K. Lindsey
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public Licence as published by
#  the Free Software Foundation; either version 2 of the Licence, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public Licence for more details.
#
#  You should have received a copy of the GNU General Public Licence
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#  SYNOPSIS
#
#	flip(dna, circular=FALSE, force.met=FALSE, min.region=0,
#		min.length=0, code=0, max.prot=100)
#
#  DESCRIPTION
#
#	Function to find ORFs in a DNA sequence

flip <- function(dna, circular=FALSE, force.met=FALSE,
	min.region=0, min.length=0, code=1, max.prot=100){
#
# C constants in flip.h
NB.FRAMES <- 6
MAX.CTX.LG <- 12
#
if(is.null(sequence))stop("a sequence to match must be supplied")
if(!is.vector(dna,mode="character"))
	stop("dna must be a character vector")
if(length(dna)>1)dna <- paste(dna,collapse="")
len <- length(strsplit(dna,split=NULL)[[1]])
str <- paste(rep(" ",trunc(len/3)),collapse="")
str2 <- paste(rep(" ",2*MAX.CTX.LG+4),collapse="")
z <- .C("flip",
	seq=as.character(dna),
	prot.nb=as.integer(max.prot),
	met=as.integer(force.met),
	circ=as.integer(circular),
	min.region=as.integer(min.region),
	min.prot=as.integer(min.length),
	code.id=as.integer(code),
	seq.lg=integer(1),
	clean.seq.lg=integer(1),
	nb.start=integer(NB.FRAMES),
	nb.stop=integer(NB.FRAMES),
	start=integer(trunc(len/3)*NB.FRAMES),
	stop=integer(trunc(len/3)*NB.FRAMES),
	prot.seq=as.character(rep(str,NB.FRAMES)),
	context.ends=integer(2*max.prot),
	context=as.character(rep(str2,max.prot)),
	protein=as.character(rep(str,max.prot)),
	error=as.integer(0))
#
# stop if error
#
if(z$error==1)stop("cannot allocate memory")
if(z$error==2)stop("minimum coding region length must be positive")
if(z$error==3)stop("minimum protein length must be positive")
if(z$error==4)stop("code id must be between 1 and 6 or 9 and 20")
if(z$error==5)stop(paste("more than",prot.nb,"proteins found"))
#
# fix up results removing empty ends of vectors
#
z$start <- matrix(z$start,ncol=6,byrow=TRUE)[1:max(z$nb.start),]+1
z$stop <- matrix(z$stop,ncol=6,byrow=TRUE)[1:max(z$nb.stop),]+1
for(i in 1:NB.FRAMES){
	if(z$nb.start[i]<max(z$nb.start))
		for(j in (z$nb.start[i]+1):max(z$nb.start))
			z$start[j,i] <- NA
	if(z$nb.stop[i]<max(z$nb.stop))
		for(j in (z$nb.stop[i]+1):max(z$nb.stop))
			z$stop[j,i] <- NA}
z$context.ends <- matrix(z$context.ends[1:(2*z$prot.nb)],ncol=2)
z$context <- z$context[1:z$prot.nb]
z$protein <- z$protein[1:z$prot.nb]
z$frame <- ifelse(z$context.ends[,2]>z$context.ends[,1],
		((z$context.ends[,1]+2)%%3)+1,-(((z$context.ends[,2]+2)%%3)+1))
#
# print results
#
cat("\nOpen reading frames found:\n")
for(i in 1:z$prot.nb){
	compl <- !z$context.ends[i,2]>z$context.ends[i,1]
	cat("\n",i,". Sequence positions ",z$context.ends[i,1]," to ",
		z$context.ends[i,2]," in ",if(compl)"complementary ",
		"frame ",abs(z$frame[i])," with length ",
		if(!compl) (z$context.ends[i,2]-z$context.ends[i,1]+1)/3
		else (z$context.ends[i,1]-z$context.ends[i,2]+1)/3,"\n",sep="")
	cat("Protein:",z$protein[i],"\n")
	cat("Context:",z$context[i],"\n")}
invisible(z)
}
