#
#  dna : A Library of Procedures for Analysis of DNA Sequences
#  Copyright (C) 2000 J.K. Lindsey
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public Licence as published by
#  the Free Software Foundation; either version 2 of the Licence, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public Licence for more details.
#
#  You should have received a copy of the GNU General Public Licence
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#  SYNOPSIS
#
#	match.DNA(dna, sequence=NULL, gaps=NULL)
#	find.DNA(dna, what="ends", gaps=NULL, position=TRUE)
#
#  DESCRIPTION
#
#	Functions for finding matches in DNA sequences

###
### Match short sequences in long sequence
###
match.DNA <- function(dna, sequence=NULL, gaps=NULL){
mp <- function(z) paste(z,collapse="")
#
# prepare sequence
#
if(is.null(sequence))stop("a sequence to match must be supplied")
if(!is.vector(dna,mode="character"))
	stop("dna must be a character vector")
if(length(dna)==1)dna <- strsplit(dna,split=NULL)[[1]]
dna <- gsub("U","T",dna)
mat <- match(dna,c(gaps,"A","C","G","T"))
if(any(is.na(mat)))
	stop(paste("dna must only contain A, C, G, T, U, or gaps\nFound",dna[is.na(mat)]))
#
# match forward
#
ldna <- length(dna)
ls <- length(strsplit(sequence,split=NULL)[[1]])
mat <- list()
for(i in 1:ls){
	ln <- trunc((ldna-i+1)/ls)*ls+i-1
	tmp <- apply(matrix(dna[i:ln],nrow=ls),2,mp)
	mm <- match(tmp,sequence)*(1:length(tmp))*ls-ls+i
	mat <- c(mat,list(mm[!is.na(mm)]))}
names(mat) <- paste("frame",1:ls,sep="")
z <- list(forward=mat)
#
# switch codes
#
dna <- complement.DNA(dna,gaps=gaps)
#
# match backwards
#
mat <- list()
for(i in 1:ls){
	ln <- trunc((ldna-i+1)/ls)*ls+i-1
	tmp <- apply(matrix(dna[ln:i],nrow=ls),2,mp)
	mm <- match(tmp,sequence)*(1:length(tmp))*ls-ls+i
	mat <- c(mat,list(mm[!is.na(mm)]))}
names(mat) <- paste("frame",1:ls,sep="")
z$backward <- mat
#
# print the results
#
fw <- bw <- FALSE
for(i in 1:ls){
	if(length(z$forward[[i]])>0)fw <- TRUE
	if(length(z$backward[[i]])>0)bw <- TRUE}
if(fw)cat("\nForward positions:")
else cat("\nNo forward positions")
for(i in 1:ls)if(length(z$forward[[i]])>0)
	cat(paste("\n Reading frame",i,"\n"),z$forward[[i]])
if(bw)cat("\nBackward positions:")
else cat("\nNo backward positions")
for(i in 1:ls)if(length(z$backward[[i]])>0)
	cat(paste("\n Reading frame",i,"\n"),z$backward[[i]])
cat("\n")
invisible(z)
}

###
### Find special code sets in a sequence
###
find.DNA <- function(dna, what="ends", gaps=NULL, position=TRUE, length=1){
mp <- function(z) paste(z,collapse="")
what <- match.arg(what,c("ends","runs"))
#
# prepare sequence
#
if(is.null(sequence))stop("a sequence to match must be supplied")
if(!is.vector(dna,mode="character"))
	stop("dna must be a character vector")
s1 <- length(dna)==1
if(s1)dna <- strsplit(dna,split=NULL)[[1]]
if(length(gaps)>0)for(i in 1:length(gaps))dna <- dna[dna!=gaps[i]]
mat <- match(dna,c("A","C","G","T"))
if(any(is.na(mat)))
	stop(paste("dna must only contain A, C, G, or T\nFound",dna[is.na(mat)]))
ldna <- length(dna)
#
# find start and stop codes
#
if(what=="ends"){
	# match forward
	mat <- list()
	for(i in 1:3){
		ln <- trunc((ldna-i+1)/3)*3+i-1
		tmp <- apply(matrix(dna[i:ln],nrow=3),2,mp)
		if(position){
			tmp2 <- list()
			mm <- match(tmp,"ATG")*(1:length(tmp))*3-3+i
			tmp2$start <- mm[!is.na(mm)]
			mm <- match(tmp,"TAA")*(1:length(tmp))*3-3+i
			tmp2$stop <- mm[!is.na(mm)]
			mm <- match(tmp,"TAG")*(1:length(tmp))*3-3+i
			tmp2$stop <- c(tmp2$stop,mm[!is.na(mm)])
			mm <- match(tmp,"TGA")*(1:length(tmp))*3-3+i
			tmp2$stop <- c(tmp2$stop,mm[!is.na(mm)])
			if(!is.null(tmp2$stop))tmp2$stop <- sort(tmp2$stop)
			tmp <- tmp2}
		else {
			tmp <- gsub("ATG","sta",tmp)
			tmp <- gsub("TAA|TAG|TGA","stp",tmp)
			if(s1)tmp <- paste(tmp,collapse="")}
		mat <- c(mat,list(tmp))}
	names(mat) <- paste("frame",1:3,sep="")
	z <- list(forward=mat)
	# switch codes
	dna <- complement.DNA(dna,gaps=gaps)
	# match backwards
	mat <- list()
	for(i in 1:3){
		ln <- trunc((ldna-i+1)/3)*3+i-1
		tmp <- apply(matrix(dna[ln:i],nrow=3),2,mp)
		if(position){
			tmp2 <- list()
			mm <- match(tmp,"ATG")*(1:length(tmp))*3-3+i
			tmp2$start <- mm[!is.na(mm)]
			mm <- match(tmp,"TAA")*(1:length(tmp))*3-3+i
			tmp2$stop <- mm[!is.na(mm)]
			mm <- match(tmp,"TAG")*(1:length(tmp))*3-3+i
			tmp2$stop <- c(tmp2$stop,mm[!is.na(mm)])
			mm <- match(tmp,"TGA")*(1:length(tmp))*3-3+i
			tmp2$stop <- c(tmp2$stop,mm[!is.na(mm)])
			if(!is.null(tmp2$stop))tmp2$stop <- sort(tmp2$stop)
			tmp <- tmp2}
		else {
			tmp <- gsub("ATG","sta",tmp)
			tmp <- gsub("TAA|TAG|TGA","stp",tmp)
			if(s1)tmp <- paste(tmp,collapse="")}
		mat <- c(mat,list(tmp))}
	names(mat) <- paste("frame",1:3,sep="")
	z$backward <- mat
	if(position){
	# print the results
	        fw <- bw <- FALSE
	        for(i in 1:3)for(j in 1:2){
	        	if(length(z$forward[[i]][[j]])>0)fw <- TRUE
	        	if(length(z$backward[[i]][[j]])>0)bw <- TRUE}
	        if(fw)cat("\nForward positions:")
	        else cat("\nNo forward positions")
	        for(i in 1:3){
	        	if(length(z$forward[[i]][[1]])>0)
	        		cat(paste("\n Reading frame",i,"start\n"),z$forward[[i]][[1]])
	        	if(length(z$forward[[i]][[2]])>0)
	        		cat(paste("\n Reading frame",i,"stop\n"),z$forward[[i]][[2]])}
	        if(bw)cat("\nBackward positions:")
	        else cat("\nNo backward positions")
	        for(i in 1:3){
	        	if(length(z$backward[[i]][[1]])>0)
	        		cat(paste("\n Reading frame",i,"start\n"),z$backward[[i]][[1]])
	        	if(length(z$backward[[i]][[2]])>0)
	        		cat(paste("\n Reading frame",i,"stop\n"),z$backward[[i]][[2]])}
	        cat("\n")}}
else if(what=="runs"){
	# forward runs
	z1 <- list()
	for(i in 1:length){
		ln <- trunc((ldna-i+1)/length)*length+i-1
		tmp <- apply(matrix(dna[i:ln],nrow=length),2,mp)
		ldna2 <- length(tmp)
		i <- c(which(tmp[1:(ldna2-1)]!=tmp[2:ldna2]),ldna2)
		values <- tmp[i]
		names(values) <- diff(c(0,i))
		z1 <- c(z1,list(list(runs=values,frequencies=table(diff(c(0,i)),tmp[i]))))}
	names(z1) <- paste("frame",1:length,sep="")
	# switch codes
	dna <- complement.DNA(dna,gaps=gaps)
	# backward runs
	z2 <- list()
	for(i in 1:length){
		ln <- trunc((ldna-i+1)/length)*length+i-1
		tmp <- apply(matrix(dna[i:ln],nrow=length),2,mp)
		ldna2 <- length(tmp)
		i <- c(which(tmp[1:(ldna2-1)]!=tmp[2:ldna2]),ldna2)
		values <- tmp[i]
		names(values) <- diff(c(0,i))
		z2 <- c(z2,list(list(runs=values,frequencies=table(diff(c(0,i)),tmp[i]))))}
	names(z2) <- paste("frame",1:length,sep="")
	z <- list(forward=z1,backward=z2)
	cat("Frequencies of runs:\n")
	for(i in 1:length){
		if(length>1)cat("\nFrame",i,"forward")
		else cat("\nForward")
		print(z$forward[[i]]$frequencies)}
	for(i in 1:length){
		if(length>1)cat("\nFrame",i,"backward")
		else cat("\nBackward")
		print(z$backward[[i]]$frequencies)}}
if(position)invisible(z) else z
}
