#
#  dna : A Library of Procedures for Analysis of DNA Sequences
#  Copyright (C) 2000 J.K. Lindsey
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public Licence as published by
#  the Free Software Foundation; either version 2 of the Licence, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public Licence for more details.
#
#  You should have received a copy of the GNU General Public Licence
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#  SYNOPSIS
#
#	dotmatrix(dna, sequences=NULL, window=1, pch=NULL)
#	plotCG(dna, window=99, gaps=NULL)
#
#  DESCRIPTION
#
#	Plotting functions for DNA sequences

###
### Plot a dot matrix comparing two DNA sequences
###
dotmatrix <- function(dna, sequences=NULL, window=1, pch=NULL){
#
# find sequences to plot
#
obj <- inherits(dna,"repeated")||inherits(dna,"response")
if(inherits(dna,"repeated")){
	nobs <- dna$response$nobs
	dna <- dna$response$y}
else if(inherits(dna,"response")){
	nobs <- dna$nobs
	dna <- dna$y}
if(obj){
	if(length(nobs)<2)stop("data object only contains one sequence")
	if(is.null(sequences)&&length(nobs)==2)sequences <- 1:2
	else if(!is.numeric(sequences)||length(sequences)!=2)
		stop("sequences must give two sequence numbers")
	if(any(sequences>length(nobs)))stop("sequence not found")
	dna1 <- dna[rep(1:length(nobs),nobs)==sequences[1]]
	dna2 <- dna[rep(1:length(nobs),nobs)==sequences[2]]
	sequences <- paste("DNA",sequences,sep="")}
else if(is.list(dna)){
	if(length(dna)!=2||!is.vector(dna[[1]])||!is.vector(dna[[2]]))
		stop("dna must be a list containing two vectors")
	sequences <- if(is.null(names(dna)))paste("DNA",1:2,sep="")
		else names(dna)
	dna1 <- dna[[1]]
	dna2 <- dna[[2]]}
else if(is.matrix(dna)){
	if(dim(dna)[1]==1)stop("Two sequences must be compared")
	else if(dim(dna)[1]>2){
		if(is.null(sequences))
			stop("Specify two names or numbers of sequences")
		if(length(sequences)!=2)stop("Two sequences must be given")
		if(is.null(rownames(dna))){
			if(!is.numeric(sequences))
				stop("sequences must give two sequence numbers")
			rown <- sequences
			sequences <- paste("DNA",rown,sep="")}
		else rown <- match(sequences,rownames(dna))
		if(any(is.na(rown)))
			stop(paste(sequences[is.na(rown)],"not found"))}
	else {
		rown <- 1:2
		sequences <- if(is.null(rownames(dna)))paste("DNA",1:2,sep="")
			else rownames(dna)}
	dna1 <- dna[rown[1],]
	dna2 <- dna[rown[2],]}
else stop("dna must be a response or repeated object, a list, or a matrix")
#
# collapse to window size
#
len <- trunc(min(c(length(dna1),length(dna2)))/window)
len2 <- len*window
if(is.null(pch))pch <- if(len>250) "." else 20
if(window>1){
	mp <- function(z) paste(z,collapse="")
	dna1 <- apply(matrix(dna1[1:len2],nrow=window),2,mp)
	dna2 <- apply(matrix(dna2[1:len2],nrow=window),2,mp)}
#
# plot points for identity between sequences
#
matches <- sum(dna1==dna2)
lm <- 0
align <- rbind(dna1,dna2)
xlim <- ylim <- length(dna1)
plot((1:len)[dna1==dna2],((len:1)*(dna1==dna2))[dna1==dna2],xlab=sequences[1],
	ylab=sequences[2],pch=pch,xlim=c(0,xlim),ylim=c(0,ylim))
for(i in 1:(len-1)){
	j <- len-i
	k <- i+1
	l <- dna1[1:j]==dna2[k:len]
	if(sum(l)>matches){
		matches <- sum(l)
		lm <- i
		align <- rbind(dna1[1:j],dna2[k:len])}
	points((1:j)[l],((j:1)*(l))[l],pch=pch)
	l <- dna1[k:len]==dna2[1:j]
	if(sum(l)>matches){
		matches <- sum(l)
		lm <- -i
		align <- rbind(dna1[k:len],dna2[1:j])}
	points((k:len)[l],((len:k)*(l))[l],pch=pch)}
invisible(list(matches=matches,lag=lm,alignment=align))
}

###
### Plot the proportion of CG content for a given window
###
plotCG <- function(dna, window=99, gaps=NULL){
#
# prepare sequence
#
if(!is.vector(dna,mode="character"))
	stop("dna must be a character vector")
if(length(dna)==1)dna <- strsplit(dna,split=NULL)[[1]]
dna <- gsub("U","T",dna)
mat <- match(dna,c(gaps,"A","C","G","T"))
if(any(is.na(mat)))
	stop(paste("dna must only contain A, C, G, T, U, or gaps\nFound",dna[is.na(mat)]))
#
# calculate proportions
#
dna <- gsub("A","T",gsub("C","G",dna))
window <- trunc(window/3)*3
ln <- trunc(length(dna)/window)*window
gp <- rep(TRUE,length(dna))
if(!is.null(gaps))for(i in 1:length(gaps))gp <- gp&dna!=gaps[i]
prop <- as.vector(table(dna[gp]))
prop <- prop[1]/(prop[1]+prop[2])
#
# all positions
#
mat <- matrix(dna[1:ln],nrow=window)
fr <- apply(mat,2,table)
if(is.list(fr)){
	tmp <- matrix(0,nrow=2,ncol=length(fr))
	j <- 0
	cn <- NULL
	for(i in fr){
		j <- j+1
		row <- match(dimnames(i)[[1]],c("G","T"))
		tmp[row[!is.na(row)],j] <- as.vector(i[!is.na(row)])}
	fr <- tmp}
pr <- fr[1,]/apply(fr,2,sum)
#
# position 1
#
mat1 <- matrix(dna[seq(1,ln,by=3)],nrow=window/3)
fr1 <- apply(mat1,2,table)
if(is.list(fr1)){
	tmp <- matrix(0,nrow=2,ncol=length(fr1))
	j <- 0
	cn <- NULL
	for(i in fr1){
		j <- j+1
		row <- match(dimnames(i)[[1]],c("G","T"))
		tmp[row[!is.na(row)],j] <- as.vector(i[!is.na(row)])}
	fr1 <- tmp}
pr1 <- fr1[1,]/apply(fr1,2,sum)
#
# position 2
#
mat2 <- matrix(dna[seq(2,ln,by=3)],nrow=window/3)
fr2 <- apply(mat2,2,table)
if(is.list(fr2)){
	tmp <- matrix(0,nrow=2,ncol=length(fr2))
	j <- 0
	cn <- NULL
	for(i in fr2){
		j <- j+1
		row <- match(dimnames(i)[[1]],c("G","T"))
		tmp[row[!is.na(row)],j] <- as.vector(i[!is.na(row)])}
	fr2 <- tmp}
pr2 <- fr2[1,]/apply(fr2,2,sum)
#
# position 3
#
mat3 <- matrix(dna[seq(3,ln,by=3)],nrow=window/3)
fr3 <- apply(mat3,2,table)
if(is.list(fr3)){
	tmp <- matrix(0,nrow=2,ncol=length(fr3))
	j <- 0
	cn <- NULL
	for(i in fr3){
		j <- j+1
		row <- match(dimnames(i)[[1]],c("G","T"))
		tmp[row[!is.na(row)],j] <- as.vector(i[!is.na(row)])}
	fr3 <- tmp}
pr3 <- fr3[1,]/apply(fr3,2,sum)
#
# plot results
#
plot(seq(window/2,ln,by=window),pr,type="l",xlim=c(0,length(dna)),ylim=c(0,1),
	xlab="Position",ylab="Proportion CG")
lines(seq(window/2,ln,by=window),pr1,lty=2)
lines(seq(window/2,ln,by=window),pr2,lty=3)
lines(seq(window/2,ln,by=window),pr3,lty=4)
lines(c(1,length(dna)),c(prop,prop))
legend(0,1,legend=c(paste("Window=",window),
	"All","Position 1","Position 2","Position 3"),lty=0:4,bty="n")
z <- rbind(pr,pr1,pr2,pr3)
dimnames(z) <- list(c("All","Pos1","Pos2","Pos3"),seq(window/2,ln,by=window))
invisible(z)
}
