#
#  dna : A Library of Procedures for Analysis of DNA Sequences
#  Copyright (C) 2000 J.K. Lindsey
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public Licence as published by
#  the Free Software Foundation; either version 2 of the Licence, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public Licence for more details.
#
#  You should have received a copy of the GNU General Public Licence
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#  SYNOPSIS
#
#	stats.DNA(dna, mitochondria=FALSE)
#	mutations(dna, gaps=NULL, group=FALSE)
#
#  DESCRIPTION
#
#	Statistical functions for summarizing DNA sequences

###
### Print elementary descriptive statistics about a coding sequence
###
stats.DNA <- function(dna, mitochondria=FALSE){
mp <- function(z) paste(z,collapse="")
#
# prepare sequence
#
if(!is.vector(dna,mode="character"))
	stop("dna must be a character vector")
if(length(dna)==1)dna <- strsplit(dna,split=NULL)[[1]]
if(trunc(length(dna)/3)!=length(dna)/3)
	stop("dna must contain complete triplets")
rna <- length(grep("U",dna))>0
if(!rna)dna <- gsub("T","U",dna)
mat <- match(dna,c("A","C","G","U"))
if(any(is.na(mat)))
	stop(paste("dna must only contain A, C, G, T, or U\nFound",dna[is.na(mat)]))
#
# frequencies of bases in triplets
#
tab1 <- NULL
for(i in 1:3){
	tmp <- table(dna[seq(i,length(dna),by=3)])
	if(length(tmp)<4){
		tmp2 <- rep(0,4)
		row <- match(rownames(tmp),c("A","C","G","U"))
		tmp2[row[!is.na(row)]] <- as.vector(tmp)
		tmp <- tmp2}
	tab1 <- rbind(tab1,as.vector(tmp))}
rownames(tab1) <- 1:3
colnames(tab1) <- if(rna) c("A","C","G","U") else c("A","C","G","T")
marg <- apply(tab1,1,sum)
tab2 <- tab1/marg
#
# calculate frequencies of amino acids
#
dna2 <- DNAtoProt(dna,mitochondria=mitochondria)
mat1 <- as.matrix(table(dna2))
mat1 <- cbind(mat1,3*mat1/length(dna))
colnames(mat1) <- c("Frequency","Proportion")
#
# calculate frequencies of ways of coding amino acids
#
dna2 <- apply(matrix(dna,nrow=3),2,mp)
mat2 <- table(dna2)
rn <- rownames(mat2)
mat2 <- cbind(matrix(mat2),3*matrix(mat2)/length(dna))
#
# match codes for labels
#
rn2 <- NULL
for(i in 1:length(rn))rn2 <- c(rn2,strsplit(rn[i],split=NULL)[[1]])
rn2 <- DNAtoProt(rn2,mitochondria=mitochondria)
if(!rna)rn <- gsub("U","T",rn)
rn <- paste(rn2,":",rn,sep="")
o <- order(rn)
dimnames(mat2) <-  list(rn,c("Frequency","Proportion"))
mat2 <- mat2[o,,drop=FALSE]
#
# print results
#
cat("\nFrequencies of bases within triplets:\n")
print.default(tab1)
cat("\nProportions of bases within triplet positions:\n")
print.default(tab2,digits=3,print.gap=1)
if(dim(mat1)[1]<dim(mat2)[1]){
	cat("\nFrequencies of amino acids:\n")
	print.default(mat1,digits=3,print.gap=1)}
cat("\nFrequencies of codes for amino acids:\n")
print.default(mat2,digits=3,print.gap=1)
invisible(list(freq.triplets=tab1,prop.triplets=tab2,amino.acids=mat1,
	amino.codes=mat2))
}

###
### Calculate frequencies of mutation in a set of sequences
###
mutations <- function(dna, gaps=NULL, group=FALSE){
#
# prepare vectors
#
if(inherits(dna,"repeated")){
	if(any(diff(dna$response$nobs)!=0))
		stop("All sequences must be the same length")
	dna <- matrix(dna$response$y,nrow=length(dna$response$nobs),byrow=TRUE)
	rn <- 0:max(dna)}
else if(inherits(dna,"response")){
	if(any(diff(dna$nobs)!=0))
		stop("All sequences must be the same length")
	dna <- matrix(dna$y,nrow=length(dna$nobs),byrow=TRUE)
	rn <- 0:max(dna)}
else if(!is.matrix(dna))
	stop("dna must be a response or repeated object or a matrix")
else rn <- c(gaps,"A","C","G","T")
if(dim(dna)[1]<3)stop("At least three sequences must be used")
#
# calculate frequencies
#
tmp1 <- apply(dna,2,table)
tmp2 <- matrix(0,nrow=length(rn),ncol=length(tmp1))
rownames(tmp2) <- rn
j <- 0
cn <- NULL
for(i in tmp1){
	j <- j+1
	cn <- c(cn,dimnames(i)[[1]][match(max(i),i)[1]])
	row <- match(dimnames(i)[[1]],rn)
	tmp2[row[!is.na(row)],j] <- as.vector(i[!is.na(row)])}
colnames(tmp2)<- cn
if(group){
	tmp1 <- apply(tmp2,2,max)
	tmp2 <- rbind(tmp1,apply(tmp2,2,sum)-tmp1)
	rownames(tmp2) <- c("stable","mutation")}
tmp2
}
